package de.hft.stuttgart.citydoctor2.math;

import Jama.Matrix;
import de.hft.stuttgart.citydoctor2.datastructure.*;
import de.hft.stuttgart.citydoctor2.utils.Copy;
import org.citygml4j.core.model.core.ImplicitGeometry;
import org.citygml4j.core.model.core.TransformationMatrix4x4;
import org.xmlobjects.gml.model.geometry.primitives.Point;
import org.xmlobjects.gml.model.geometry.primitives.PointProperty;
import org.xmlobjects.util.copy.CopyBuilder;

import java.util.*;

/**
 * Adapter class for CityGML-TransformationMatrix4x4
 */
public class TransformationMatrix {

    private final TransformationMatrix4x4 originalTM4x4;

    private Matrix tMatrix;
    /**
     * The {@code TransformationMatrix} class is an adapter for applying the {@link TransformationMatrix4x4} of implicitGeometries to geometries
     * of CityDoctor's internal model.
     * @param transformationMatrix the transformation matrix.
     * @param pointProperty the reference point for the translation of the geometry's position.
     */
    public TransformationMatrix(TransformationMatrix4x4 transformationMatrix, PointProperty pointProperty) {
        originalTM4x4 = transformationMatrix;
        setupMatrix();
        incorporateRefPoint(pointProperty);
    }

    /**
     * The {@code TransformationMatrix} class is an adapter for applying the {@link TransformationMatrix4x4} of implicitGeometries to geometries
     * of CityDoctor's internal model.
     * @param implicitGeometry the implicitGeometry containing the TransformationMatrix and ReferencePoint.
     */
    public TransformationMatrix(ImplicitGeometry implicitGeometry) {
        originalTM4x4 = implicitGeometry.getTransformationMatrix();
        setupMatrix();
        incorporateRefPoint(implicitGeometry.getReferencePoint());

    }


    private void setupMatrix() {
        double[][] matrix = new double[4][4];
        CopyBuilder copyBuilder = new CopyBuilder();
        TransformationMatrix4x4 tmMatrix = (TransformationMatrix4x4) originalTM4x4.deepCopy(copyBuilder);
        List<Double> tmValues = tmMatrix.toRowMajorList();
        for (int i = 0; i < 4 ; i++){
            for (int j = 0; j < 4 ; j++){
                matrix[i][j] = tmValues.get(i*4+j);
            }
        }
        tMatrix = new Matrix(matrix);
    }

    private void incorporateRefPoint(PointProperty pointProperty) {
        Point refPoint = pointProperty.getObject();
        List<Double> refCoord =  refPoint.getPos().toCoordinateList3D();
        double [][] tArray = tMatrix.getArray();
        tArray[0][3] = refCoord.get(0);
        tArray[1][3] = refCoord.get(1);
        tArray[2][3] = refCoord.get(2);
    }


    /**
     * Transforms a vector with the transformation matrix.
     * @param referenceGeom reference geometry to transform.
     * @return A transformed copy of {@code referenceGeom}.
     */
    public Geometry transformGeometry(Geometry referenceGeom) {

        // Create deep copy of reference geom
        Geometry geom = Copy.copy(referenceGeom);
        Map<Vertex, Boolean> visitedVertices = new IdentityHashMap<>();
        Set<Vertex> visitedVerticesSet = Collections.newSetFromMap(visitedVertices);
        // Transform all vertices
        for (Polygon polygon : geom.getPolygons()){
            for (Vertex vertex : polygon.getExteriorRing().getVertices()){
                if (!visitedVerticesSet.contains(vertex)){
                    transformVector3D(vertex);
                    visitedVerticesSet.add(vertex);
                }
            }
            for (LinearRing ring : polygon.getInnerRings()){
                for (Vertex vertex : ring.getVertices()){
                    if (!visitedVerticesSet.contains(vertex)){
                        transformVector3D(vertex);
                        visitedVerticesSet.add(vertex);
                    }
                }
            }
        }
        geom.updateEdgesAndVertices();
        return geom;
    }


    public void transformVector3D(Vector3d vector) {
        double[] tfVector = new double[]{vector.getX(), vector.getY(), vector.getZ(), 1d};
        Matrix vectorMatrix = new Matrix(tfVector,4);
        Matrix res = tMatrix.times(vectorMatrix);
        double[] resultVector = res.getRowPackedCopy();
        vector.setX(resultVector[0]);
        vector.setY(resultVector[1]);
        vector.setZ(resultVector[2]);
    }



}
