package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.CityDoctorValidation;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.datastructure.BoundingBox;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.FeatureType;
import de.hft.stuttgart.citydoctor2.gui.logger.GuiLogger;
import de.hft.stuttgart.citydoctor2.gui.tree.FeatureNode;
import de.hft.stuttgart.citydoctor2.gui.tree.Renderable;
import de.hft.stuttgart.citydoctor2.gui.tree.RenderableTreeCell;
import de.hft.stuttgart.citydoctor2.parameter.ArgumentParser;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ProgressListener;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.embed.swing.SwingFXUtils;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.*;
import javafx.scene.control.*;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.image.WritableImage;
import javafx.scene.input.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.Priority;
import javafx.scene.paint.Color;
import javafx.scene.transform.Rotate;
import javafx.stage.Stage;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.*;

import static javafx.scene.input.Clipboard.getSystemClipboard;

public class MainWindow extends Application {

    private static final Logger logger = LogManager.getLogger(MainWindow.class);

    private static final double CAMERA_TRANSLATE_Z = -100.0;
    private static final double CAMERA_INITIAL_X_ANGLE = 20.0;
    private static final double CAMERA_INITIAL_Y_ANGLE = 120.0;

    @FXML
    private TreeView<Renderable> buildingsView;

    @FXML
    private TreeView<Renderable> vegetationView;

    @FXML
    private TreeView<Renderable> transView;

    @FXML
    private TreeView<Renderable> tunnelView;

    @FXML
    private TreeView<Renderable> bridgeView;

    @FXML
    private TreeView<Renderable> waterView;

    @FXML
    private TreeView<Renderable> terrainView;

    @FXML
    private TreeView<Renderable> cityFurnitureView;

    @FXML
    private TreeView<Renderable> attributeView;

    @FXML
    TreeView<Renderable> otherObjectsView;

    @FXML
    private TreeView<Renderable> polygonView;

    @FXML
    private TreeView<Renderable> edgeView;

    @FXML
    private TreeView<Renderable> vertexView;

    @FXML
    private Pane meshView;

    @FXML
    private SplitPane mainContainer;

    @FXML
    private TabPane detailsTabPane;

    @FXML
    private TabPane featurePane;

    @FXML
    private TextField searchField;

    @FXML
    private TreeView<Renderable> errorView;

    @FXML
    private Button clearBtn;

    @FXML
    private Button searchBtn;

    @FXML
    private ToggleButton allButton;

    @FXML
    private ToggleButton errorButton;

    @FXML
    private ComboBox<String> showCityObjectsCombo;

    @FXML
    private HBox viewPane;

    @FXML
    private ListView<String> globalErrorsView;

    @FXML
    private BorderPane mainPane;

    @FXML
    private HBox viewButtonBox;

    @FXML
    private ComboBox<Locale> languageSelector;

    @FXML
    private Tab buildingsTab;

    @FXML
    private Tab vegetationTab;

    @FXML
    private Tab transportationTab;

    @FXML
    private Tab tunnelTab;

    @FXML
    private Tab bridgeTab;

    @FXML
    private Tab waterTab;

    @FXML
    private Tab terrainTab;

    @FXML
    private Tab cityFurnitureTab;

    @FXML
    private Tab otherObjectsTab;

    @FXML
    private Tab errorsTab;

    @FXML
    private Tab polygonsTab;

    @FXML
    private Tab edgesTab;

    @FXML
    private Tab verticesTab;

    @FXML
    private Tab attributeTab;

    @FXML
    private Tab logTab;

    @FXML
    private Tab globalErrorsTab;

    @FXML
    private Label viewLabel;

    @FXML
    private Label showLabel;

    @FXML
    private Label searchLabel;

    @FXML
    private Label memoryLabel;

    @FXML
    private ProgressBar memoryBar;

    @FXML
    private Label memoryConsumptionLabel;

    @FXML
    private Label availableLabel;

    private Image north;
    private Image northFlip;

    private Group meshGroup;

    private Group world;
    private PerspectiveCamera camera;
    private Rotate cameraXRotation = new Rotate();
    private Rotate cameraZRotation = new Rotate();

    private double dragX;
    private double dragY;

    private double cameraXRot = CAMERA_INITIAL_X_ANGLE;
    private double cameraYRot = CAMERA_INITIAL_Y_ANGLE;
    private double translateZ = CAMERA_TRANSLATE_Z;
    private double[] clickStart = new double[2];

    @FXML
    private TextArea logArea;

    @FXML
    private ToolBar viewBar;

    private Stage stage;
    private ExceptionDialog exceptionDialog;

    private Renderer renderer;
    private CityDoctorController controller;
    private HighlightController highlightController;

    private FeatureType selectedTab = FeatureType.BUILDING;

    private MainToolBar mainToolBar;
    private SubScene geomScene;
    private ImageView northArrow;

    private static boolean loadFileAtStartup = false;
    private static String inputFile;
    private static String xmlOutput;
    private static String pdfOutput;
    private static ValidationConfiguration config;

    private VertexClickHandler clickHandler;

    private ChangeListener<Number> filterChangeListener;

    public static void main(String[] args) {
        setLocaleFromSettings();
        ArgumentParser argParser = new ArgumentParser(args);
        inputFile = CityDoctorValidation.getInputFile(argParser, true);
        if (inputFile != null) {
            loadFileAtStartup = true;
            xmlOutput = CityDoctorValidation.getXmlOutput(argParser);
            pdfOutput = CityDoctorValidation.getPdfOutput(argParser);
            try {
                config = CityDoctorValidation.getValidationConfig(argParser, true);
            } catch (FileNotFoundException e) {
                Platform.runLater(() -> {
                    List<String> configFiles = argParser.getValues("config");
                    Alert alert = new Alert(AlertType.ERROR);
                    alert.setContentText(Localization.getText("MainWindow.missingConfig") + configFiles);
                    alert.showAndWait();
                });
                System.exit(4);
            }
        }
        Application.launch(args);
    }

    private static void setLocaleFromSettings() {
        String localeString = Settings.get(Settings.LANGUAGE);
        if (localeString != null) {
            try {
                Locale loc = new Locale(localeString);
                Locale.setDefault(loc);
            } catch (Exception e) {
                logger.warn("Could not set language to {}, using system language", localeString);
            }
        }
    }

    @Override
    public void start(Stage stage) throws IOException {
        this.stage = stage;
        stage.getIcons().add(new Image(MainWindow.class.getResourceAsStream("icons/citydoctor_logo.png")));
        FXMLLoader loader = new FXMLLoader(MainWindow.class.getResource("MainWindow.fxml"));
        loader.setController(this);
        BorderPane bp = loader.load();
        highlightController = new HighlightController(world);
        renderer = new Renderer(this, highlightController);
        clickHandler = new VertexClickHandler(errorView, renderer, stage, this);
        controller = new CityDoctorController(this, highlightController, renderer);
        mainToolBar = new MainToolBar(stage, controller, featurePane, renderer, this);
        viewPane.getChildren().add(mainToolBar.getToolBar());
        HBox.setHgrow(mainToolBar.getToolBar(), Priority.ALWAYS);

        ValidationView valView = new ValidationView(this, controller);
        ViewRegistration.registerView(valView);
        setupViews(valView);

        createLanguageSelector();
        setLabelsInCorrectLanguage();

        Scene scene = new Scene(bp, 1280, 800);
        createDropTarget(scene, valView);
        String version = Localization.getText(Localization.VERSION);
        stage.setTitle("CityDoctor " + version);
        stage.setScene(scene);
        stage.show();
        checkForStartupLoading();

        memoryBar.setOnMouseClicked(me -> System.gc());
        Timer timer = new Timer(true);
        // check memory every second
        TimerTask task = new TimerTask() {

            Runtime runtime = Runtime.getRuntime();

            @Override
            public void run() {
                long totalMemory = runtime.totalMemory();
                long freeMemory = runtime.freeMemory();
                long usedMemory = totalMemory - freeMemory;
                double percentage = usedMemory / (double) totalMemory;
                if (totalMemory / 1024 / 1024 >= 1024) {
                    // gb
                    double totalMemoryGb = totalMemory / (1024d * 1024d * 1024d);
                    double usedMemoryGb = usedMemory / (1024d * 1024d * 1024d);
                    String memoryString = String.format("%.1f GB / %.1f GB", usedMemoryGb, totalMemoryGb);
                    Platform.runLater(() -> {
                        memoryConsumptionLabel.setText(memoryString);
                        memoryBar.setProgress(percentage);
                    });
                } else if (totalMemory / 1024 >= 1024) {
                    // mb
                    double totalMemoryMb = totalMemory / (1024d * 1024d);
                    double usedMemoryMb = usedMemory / (1024d * 1024d);
                    String memoryString = String.format("%.1f MB / %.1f MB", usedMemoryMb, totalMemoryMb);
                    Platform.runLater(() -> {
                        memoryConsumptionLabel.setText(memoryString);
                        memoryBar.setProgress(percentage);
                    });
                }
            }
        };
        timer.schedule(task, 0, 1000);
    }

    private void createDropTarget(Scene scene, ValidationView valView) {
        setDragOverInteraction(scene, valView);
        scene.setOnDragDropped(event -> {
            if (ViewRegistration.getCurrentActiveView() != valView) {
                return;
            }
            Dragboard db = event.getDragboard();
            boolean success = false;
            if (db.hasFiles() && db.getFiles().size() == 1) {
                File f = db.getFiles().get(0);
                Thread t = new Thread(() -> {
                    try {
                        controller.loadCityGml(f.getAbsolutePath(), 8, (ProgressListener) null, false);
                    } catch (Exception e) {
                        if (logger.isErrorEnabled()) {
                            logger.error(Localization.getText("OpenFileDialog.loadFailed"), e);
                        }
                        Platform.runLater(() -> {
                            ExceptionDialog exDialog = new ExceptionDialog();
                            exDialog.show(e);
                        });
                    }
                });
                t.start();
                success = true;
            }
            // let the source know whether the string was successfully transferred and used
            event.setDropCompleted(success);

            event.consume();
        });
    }

    private void setDragOverInteraction(Scene scene, ValidationView valView) {
        scene.setOnDragOver(event -> {
            if (ViewRegistration.getCurrentActiveView() != valView) {
                return;
            }
            if (event.getGestureSource() != scene && event.getDragboard().hasFiles()
                    && event.getDragboard().getFiles().size() == 1) {
                // allow for both copying and moving, whatever user chooses
                event.acceptTransferModes(TransferMode.LINK);
            }
            event.consume();
        });
    }

    private void setLabelsInCorrectLanguage() {
        buildingsTab.setText(Localization.getText("MainWindow.buildingsTab"));
        vegetationTab.setText(Localization.getText("MainWindow.vegetationTab"));
        transportationTab.setText(Localization.getText("MainWindow.transportationTab"));
        tunnelTab.setText(Localization.getText("MainWindow.tunnelTab"));
        bridgeTab.setText(Localization.getText("MainWindow.bridgeTab"));
        waterTab.setText(Localization.getText("MainWindow.waterTab"));
        terrainTab.setText(Localization.getText("MainWindow.terrainTab"));
        cityFurnitureTab.setText(Localization.getText("MainWindow.cityfurnitureTab"));
        otherObjectsTab.setText(Localization.getText("MainWindow.otherObjectsTab"));
        viewLabel.setText(Localization.getText("MainWindow.viewLabel"));
        showLabel.setText(Localization.getText("MainWindow.showLabel"));
        searchLabel.setText(Localization.getText("MainWindow.searchLabel"));
        searchBtn.setText(Localization.getText("MainWindow.searchBtn"));
        clearBtn.setText(Localization.getText("MainWindow.clearBtn"));
        errorsTab.setText(Localization.getText("MainWindow.errorsTab"));
        polygonsTab.setText(Localization.getText("MainWindow.polygonsTab"));
        edgesTab.setText(Localization.getText("MainWindow.edgesTab"));
        verticesTab.setText(Localization.getText("MainWindow.verticesTab"));
        attributeTab.setText(Localization.getText("MainWindow.attributeTab"));
        logTab.setText(Localization.getText("MainWindow.logTab"));
        globalErrorsTab.setText(Localization.getText("MainWindow.globalErrorsTab"));
        memoryLabel.setText(Localization.getText("MainWindow.memoryLabel"));
        availableLabel.setText(String.format("%s %.1f GB", Localization.getText("MainWindow.availableLabel"),
                Runtime.getRuntime().maxMemory() / 1024d / 1024d / 1024d));
    }

    private void createLanguageSelector() {
        languageSelector.setButtonCell(new LanguageSelectorCell());
        languageSelector.setCellFactory(view -> new LanguageSelectorCell());
        languageSelector.getItems().add(Locale.GERMAN);
        languageSelector.getItems().add(Locale.ENGLISH);
        if (Locale.getDefault().getLanguage().equals(Locale.GERMAN.getLanguage())) {
            languageSelector.getSelectionModel().select(Locale.GERMAN);
        } else if (Locale.getDefault().getLanguage().equals(Locale.ENGLISH.getLanguage())) {
            languageSelector.getSelectionModel().select(Locale.ENGLISH);
        } else {
            languageSelector.getSelectionModel().select(Locale.ENGLISH);
            Settings.set(Settings.LANGUAGE, Locale.ENGLISH.getLanguage());
        }
        languageSelector.getSelectionModel().selectedItemProperty().addListener((obs, oldV, newV) -> {
            Alert alert = new Alert(AlertType.CONFIRMATION, Localization.getText("MainWindow.languageChange"),
                    ButtonType.OK);
            alert.showAndWait();
            Settings.set(Settings.LANGUAGE, newV.getLanguage());
        });
    }

    private void checkForStartupLoading() {
        if (loadFileAtStartup) {
            logger.info(Localization.getText("MainWindow.loadGivenFile"));
            Thread t = new Thread(() -> {
                try {
                    // wait a bit for the gui to show
                    Thread.sleep(500);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
                try {
                    controller.loadCityGml(inputFile, config.getNumberOfRoundingPlaces(), null,
                            config.isXmlValidation());
                    logger.info(Localization.getText("MainWindow.finishedLoading"));
                    logger.info(Localization.getText("MainWindow.checking"));
                    controller.startChecks(config, null);
                    if (xmlOutput != null) {
                        logger.info(Localization.getText("MainWindow.writeXml"));
                        controller.writeXmlReport(new File(xmlOutput));
                        logger.info(Localization.getText("MainWindow.finishedXml"));
                    }
                    if (pdfOutput != null) {
                        logger.info(Localization.getText("MainWindow.writePdf"));
                        controller.writePdfReport(new File(pdfOutput));
                        logger.info(Localization.getText("MainWindow.finishedPdf"));
                    }
                } catch (CityGmlParseException | InvalidGmlFileException e) {
                    logger.error(Localization.getText("MainWindow.loadFailed"), e.getMessage());
                }
            });
            t.start();
        }
    }

    private void setupViews(ValidationView valView) {
        ToggleGroup group = new ToggleGroup();
        for (View v : ViewRegistration.getRegisteredViews()) {

            RadioButton radioButton = new RadioButton();
            radioButton.getStyleClass().remove("radio-button");
            radioButton.getStyleClass().add("toggle-button");
            ImageView view = new ImageView(v.getViewLogo());
            view.setFitHeight(32);
            view.setFitWidth(32);
            radioButton.setGraphic(view);
            group.getToggles().add(radioButton);
            if (v == valView) {
                ViewRegistration.setCurrentActiveView(valView);
                group.selectToggle(radioButton);
            }
            viewButtonBox.getChildren().add(radioButton);
            radioButton.selectedProperty().addListener((obs, oldV, newV) -> {
                if (Boolean.TRUE.equals(newV)) {
                    showView(v);
                    ViewRegistration.setCurrentActiveView(v);
                } else {
                    v.onHide();
                }
            });
        }
    }

    private void showView(View v) {
        viewPane.getChildren().clear();
        controller.showView(v);
        Optional<HBox> toolbar = v.getToolbar();
        if (toolbar.isPresent()) {
            viewPane.getChildren().add(toolbar.get());
            HBox.setHgrow(toolbar.get(), Priority.ALWAYS);
        } else {
            HBox placeHolderToolbar = new HBox();
            placeHolderToolbar.setMaxHeight(Double.MAX_VALUE);
            viewPane.getChildren().add(placeHolderToolbar);
            HBox.setHgrow(placeHolderToolbar, Priority.ALWAYS);
        }
        mainPane.setCenter(v.getMainScreen());
    }

    public void initialize() {
        GuiLogger.setTextArea(logArea);

        loadFrameConfig();

        setup3dView();
        setupTrees();

        setupSearchField();
        setupSearchButtons();

        setupShowCityComboBox();

        detailsTabPane.getSelectionModel().selectedIndexProperty()
                .addListener((ov, oldI, newI) -> Platform.runLater(() -> {
                    if (newI.intValue() == 0 && errorView.getSelectionModel().getSelectedItem() != null) {
                        // the first tab is selected, meaning the error tab
                        // redisplay the selected error
                        errorView.getSelectionModel().getSelectedItem().getValue().visit(renderer);
                    }
                }));
    }

    private void setupShowCityComboBox() {
        showCityObjectsCombo.getItems().addAll(Localization.getText("MainWindow.all"),
                Localization.getText("MainWindow.withErrors"));
        showCityObjectsCombo.getSelectionModel().selectFirst();

        showCityObjectsCombo.setCellFactory(param -> new ListCell<>() {
            @Override
            public void updateItem(String item, boolean empty) {
                super.updateItem(item, empty);
                if (item != null) {
                    setText(item);
                    if (item.contentEquals(Localization.getText("MainWindow.withErrors"))) {
                        setTextFill(Color.RED);
                    }
                } else {
                    setText(null);
                }
            }
        });

        filterChangeListener = setupFilterSelectionListener();
        showCityObjectsCombo.getSelectionModel().selectedIndexProperty().addListener(filterChangeListener);
    }

    public void resetFilterComboBox() {
        showCityObjectsCombo.getSelectionModel().selectedIndexProperty().removeListener(filterChangeListener);
        showCityObjectsCombo.getSelectionModel().selectFirst();
        showCityObjectsCombo.getSelectionModel().selectedIndexProperty().addListener(filterChangeListener);
    }

    private ChangeListener<Number> setupFilterSelectionListener() {
        return (obs, oldV, newV) -> controller.errorFilterIndexChanged(newV);
    }

    private void setupSearchButtons() {
        searchBtn.setOnAction(ae -> controller.searchFeature(searchField.getText(), selectedTab));
        clearBtn.setOnAction(ae -> {
            if (searchField.getText().isEmpty()) {
                // do not reset search if nothing has bee searched
                return;
            }
            controller.resetSearch(selectedTab);
            searchField.setText("");
        });
    }

    private void setupSearchField() {
        searchField.textProperty().addListener((obs, oldV, newV) -> {
            if (newV.isEmpty()) {
                controller.resetSearch(selectedTab);
            }
        });
        featurePane.getSelectionModel().selectedIndexProperty().addListener((obs, oldV, newV) -> {
            if (!searchField.getText().isEmpty()) {
                resetSearchBar();
                controller.resetSearch(selectedTab);
            }
            int index = newV.intValue();
            switch (index) {
                case 0:
                    selectedTab = FeatureType.BUILDING;
                    break;
                case 1:
                    selectedTab = FeatureType.VEGETATION;
                    break;
                case 2:
                    selectedTab = FeatureType.TRANSPORTATION;
                    break;
                case 3:
                    selectedTab = FeatureType.TUNNEL;
                    break;
                case 4:
                    selectedTab = FeatureType.BRIDGE;
                    break;
                case 5:
                    selectedTab = FeatureType.WATER;
                    break;
                case 6:
                    selectedTab = FeatureType.LAND;
                    break;
                case 7:
                    selectedTab = FeatureType.CITY_FURNITURE;
                    break;
                case 8:
                    selectedTab = FeatureType.GENERIC_CITY_OBJECT;
                    break;
                default:
                    throw new IllegalStateException("Unknown tab index: " + index);
            }
        });
    }

    public void setDisableOfFeatureTab(FeatureType featureType, boolean value) {
        switch (featureType) {
            case BUILDING:
                buildingsTab.setDisable(value);
                break;
            case VEGETATION:
                vegetationTab.setDisable(value);
                break;
            case TRANSPORTATION:
                transportationTab.setDisable(value);
                break;
            case TUNNEL:
                tunnelTab.setDisable(value);
                break;
            case BRIDGE:
                bridgeTab.setDisable(value);
                break;
            case WATER:
                waterTab.setDisable(value);
                break;
            case LAND:
                terrainTab.setDisable(value);
                break;
            case CITY_FURNITURE:
                cityFurnitureTab.setDisable(value);
                break;
            case GENERIC_CITY_OBJECT:
                otherObjectsTab.setDisable(value);
                break;
            default:
                throw new IllegalArgumentException("Unknown feature type: " + featureType);

        }
    }

    public void activateAllTabs() {
        buildingsTab.setDisable(false);
        vegetationTab.setDisable(false);
        transportationTab.setDisable(false);
        tunnelTab.setDisable(false);
        bridgeTab.setDisable(false);
        waterTab.setDisable(false);
        terrainTab.setDisable(false);
        cityFurnitureTab.setDisable(false);
        otherObjectsTab.setDisable(false);

    }

    public void resetSearchBar() {
        searchField.setText("");
    }

    private void loadFrameConfig() {
        stage.setMaximized(Boolean.valueOf(Settings.get(Settings.MAXIMIZED, Boolean.FALSE.toString())));
        stage.maximizedProperty()
                .addListener((obs, oldV, newV) -> Settings.set(Settings.MAXIMIZED, Boolean.toString(newV)));

        String widthString = Settings.get(Settings.FRAME_WIDTH);
        if (widthString != null) {
            stage.setWidth(Double.parseDouble(widthString));
        }
        stage.widthProperty().addListener(
                (obs, oldV, newV) -> Settings.set(Settings.FRAME_WIDTH, Double.toString(newV.doubleValue())));

        String heightString = Settings.get(Settings.FRAME_HEIGHT);
        if (heightString != null) {
            stage.setHeight(Double.parseDouble(heightString));
        }
        stage.heightProperty().addListener(
                (obs, oldV, newV) -> Settings.set(Settings.FRAME_HEIGHT, Double.toString(newV.doubleValue())));
    }

    public void showExceptionDialog(Throwable e) {
        if (exceptionDialog == null) {
            exceptionDialog = new ExceptionDialog();
        }
        exceptionDialog.show(e);
    }

    private void setupTrees() {

        setupTreeRoot(errorView);
        setupSelectListener(errorView);

        setupTreeRoot(buildingsView);
        setupSelectListener(buildingsView);
        setupFeatureTabContextMenu(buildingsView);

        setupTreeRoot(vegetationView);
        setupSelectListener(vegetationView);
        setupFeatureTabContextMenu(vegetationView);

        setupTreeRoot(transView);
        setupSelectListener(transView);
        setupFeatureTabContextMenu(transView);

        setupTreeRoot(tunnelView);
        setupSelectListener(tunnelView);
        setupFeatureTabContextMenu(tunnelView);

        setupTreeRoot(bridgeView);
        setupSelectListener(bridgeView);
        setupFeatureTabContextMenu(bridgeView);

        setupTreeRoot(waterView);
        setupSelectListener(waterView);
        setupFeatureTabContextMenu(waterView);

        setupTreeRoot(terrainView);
        setupSelectListener(terrainView);
        setupFeatureTabContextMenu(terrainView);

        setupTreeRoot(cityFurnitureView);
        setupSelectListener(cityFurnitureView);
        setupFeatureTabContextMenu(cityFurnitureView);

        setupTreeRoot(otherObjectsView);
        setupSelectListener(otherObjectsView);
        setupFeatureTabContextMenu(otherObjectsView);

        setupTreeRoot(vertexView);
        setupSelectListener(vertexView);

        setupTreeRoot(polygonView);
        setupSelectListener(polygonView);

        setupTreeRoot(edgeView);
        setupSelectListener(edgeView);

        setupTreeRoot(attributeView);
        setupSelectListener(attributeView);
    }


    private void setupTreeRoot(TreeView<Renderable> tView){
        tView.setRoot(new TreeItem<>());
        tView.setCellFactory(param -> new RenderableTreeCell());
    }

    private void setupFeatureTabContextMenu(TreeView<Renderable> featureView) {
        ContextMenu menu = new ContextMenu();
        MenuItem mi = new MenuItem(Localization.getText("MainWindow.export"));
        mi.setOnAction(ea -> {
            Renderable render = featureView.getSelectionModel().getSelectedItem().getValue();
            if (render instanceof FeatureNode node) {
                try {
                    controller.export(node.getFeature());
                    logger.info(Localization.getText("CityDoctorController.export"));
                } catch (Exception e) {
                    logger.error("Export failed: Unexpected error occurred");
                    logger.error(e);
                }
            } else {
                logger.info(Localization.getText("MainWindow.infoTopLevelExport"));
            }
        });
        menu.getItems().add(mi);

        MenuItem clipMi = new MenuItem(Localization.getText("MainWindow.copyId"));
        clipMi.setOnAction(ea -> {
            Renderable render = featureView.getSelectionModel().getSelectedItem().getValue();
            if (render instanceof FeatureNode node) {
                Clipboard clipboard = getSystemClipboard();
                ClipboardContent content = new ClipboardContent();
                content.putString(node.getFeature().getGmlId().toString());
                if (!clipboard.setContent(content)) {
                    logger.warn(Localization.getText("MainWindow.warnCopyIdFailed"));
                }
            } else {
                logger.info(Localization.getText("MainWindow.infoTopLevelCopyID"));
            }

        });
        menu.getItems().add(clipMi);

        MenuItem deleteMi = new MenuItem(Localization.getText("MainWindow.delete"));
        deleteMi.setOnAction(ae -> controller.delete(featureView.getSelectionModel().getSelectedItem()));
        menu.getItems().add(deleteMi);
        featureView.setContextMenu(menu);
    }

    private void setupSelectListener(TreeView<Renderable> view) {
        view.getSelectionModel().selectedItemProperty().addListener((obs, oldI, newI) -> {
            if (newI != null) {
                newI.getValue().visit(renderer);
            }
        });
    }

    private void setup3dView() {
        Group root = new Group();
        geomScene = new SubScene(root, 500, 300, true, SceneAntialiasing.BALANCED);
        geomScene.heightProperty().bind(meshView.heightProperty());
        geomScene.widthProperty().bind(meshView.widthProperty());
        geomScene.setFill(Color.AZURE);
        meshView.getChildren().add(geomScene);

        geomScene.addEventFilter(MouseEvent.MOUSE_PRESSED, me -> {
            clickStart[0] = me.getScreenX();
            clickStart[1] = me.getScreenY();
        });

        geomScene.addEventFilter(MouseEvent.MOUSE_RELEASED, me -> {
            if (Math.abs(clickStart[0] - me.getScreenX()) > 3 || Math.abs(clickStart[1] - me.getScreenY()) > 3) {
                // skip when mouse moved too much
                return;
            }
            Node node = me.getPickResult().getIntersectedNode();
            if (node != null) {
                Object o = node.getUserData();
                if (o instanceof ClickDispatcher cd) {
                    cd.click(me, clickHandler);
                }
            }
        });

        world = new Group();
        root.getChildren().add(world);
        meshGroup = new Group();
        world.getChildren().add(meshGroup);

        north = new Image(getClass().getResourceAsStream("icons/north.png"), 50, 50, true, true);
        northFlip = new Image(getClass().getResourceAsStream("icons/north_flip.png"), 50, 50, true, true);
        northArrow = new ImageView(north);
        northArrow.setFitHeight(50);
        northArrow.setFitWidth(50);
        northArrow.setX(10);
        northArrow.setY(10);
        meshView.getChildren().add(northArrow);
        northArrow.setVisible(false);

        AmbientLight al = new AmbientLight(Color.WHITE);
        root.getChildren().add(al);
        buildCamera();
        cameraXRotation.setAxis(Rotate.X_AXIS);
        cameraZRotation.setAxis(Rotate.Z_AXIS);
        world.getTransforms().add(cameraXRotation);
        world.getTransforms().add(cameraZRotation);
        root.getChildren().add(camera);
        geomScene.setCamera(camera);

        setupMeshViewControls();
    }

    private void setupMeshViewControls() {
        meshView.setOnMousePressed(me -> {
            if (me.getButton() == MouseButton.PRIMARY) {
                dragX = me.getScreenX();
                dragY = me.getScreenY();
            }
            if (me.getButton() == MouseButton.SECONDARY) {
                dragX = me.getScreenX();
                dragY = me.getScreenY();
            }
        });

        meshView.setOnScroll(se -> {
            if (se.getDeltaY() < 0) {
                translateZ += translateZ * 0.05;
            } else {
                translateZ -= translateZ * 0.05;
            }
            camera.setTranslateZ(translateZ);
            highlightController.changeScaling(translateZ);
        });

        meshView.setOnMouseDragged(me -> {
            if (me.getButton() == MouseButton.PRIMARY) {
                double deltaX = me.getScreenX() - dragX;
                double deltaY = me.getScreenY() - dragY;
                dragX = me.getScreenX();
                dragY = me.getScreenY();

                cameraXRot += (deltaX / 3d) % 360;
                cameraYRot += (deltaY / 3d) % 360;

                cameraZRotation.setAngle(cameraXRot);
                cameraXRotation.setAngle(cameraYRot);
                alignNorthArrow();
            }
            if (me.getButton() == MouseButton.SECONDARY) {
                double translationSpeed = Math.abs(camera.getTranslateZ()) / 1000;
                double deltaX = me.getScreenX() - dragX;
                double deltaY = me.getScreenY() - dragY;
                dragX = me.getScreenX();
                dragY = me.getScreenY();

                camera.setTranslateX(camera.getTranslateX() - deltaX * translationSpeed);
                camera.setTranslateY(camera.getTranslateY() - deltaY * translationSpeed);
            }
        });
    }

    private void buildCamera() {
        camera = new PerspectiveCamera(true);
        camera.setNearClip(0.1);
        camera.setFarClip(10000d);
        camera.setTranslateZ(translateZ);
        cameraZRotation.setAngle(cameraXRot);
        cameraXRotation.setAngle(cameraYRot);
    }

    public void alignNorthArrow() {
        double absRotZ = cameraZRotation.getAngle() - 360 * Math.floor(cameraZRotation.getAngle() / 360);
        double absRotX = cameraXRotation.getAngle() - 360 * Math.floor(cameraXRotation.getAngle() / 360);

        double northRot;
        if (80 <= absRotX && absRotX <= 260) {
            northRot = 360 - absRotZ;
            if (northArrow.getImage() != north) {
                northArrow.setImage(north);
            }
        } else {
            //Camera is below horizon line -> flip north arrow horizontally
            northRot = (absRotZ + 180) % 360;
            if (northArrow.getImage() != northFlip) {
                northArrow.setImage(northFlip);
            }
        }
        northArrow.setRotate(northRot);
    }

    public void resetCamera() {
        cameraXRot = CAMERA_INITIAL_X_ANGLE;
        cameraYRot = CAMERA_INITIAL_Y_ANGLE;
        cameraZRotation.setAngle(cameraXRot);
        cameraXRotation.setAngle(cameraYRot);
        alignNorthArrow();
        camera.setTranslateX(0);
        camera.setTranslateY(0);
        if (controller.getOriginBB() != null) {
            zoomOutForBoundingBox(controller.getOriginBB());
        }
    }

    public void addFileNameToTitle(String fileName) {
        String version = Localization.getText(Localization.VERSION);
        stage.setTitle("CityDoctor " + version + " - " + fileName);
    }

    public TreeView<Renderable> getBuildingsView() {
        return buildingsView;
    }

    public TreeView<Renderable> getVegetationView() {
        return vegetationView;
    }

    public TreeView<Renderable> getTransportationView() {
        return transView;
    }

    public TreeView<Renderable> getTunnelView() {
        return tunnelView;
    }

    public TreeView<Renderable> getBridgeView() {
        return bridgeView;
    }

    public TreeView<Renderable> getWaterView() {
        return waterView;
    }

    public TreeView<Renderable> getTerrainView() {
        return terrainView;
    }

    public TreeView<Renderable> getCityFurnitureView() {
        return cityFurnitureView;
    }

    public TreeView<Renderable> getAttributeView() {
        return attributeView;
    }

    public TreeView<Renderable> getOtherObjectsView() {
        return otherObjectsView;
    }

    public TreeView<Renderable> getPolygonsView() {
        return polygonView;
    }

    public TreeView<Renderable> getEdgeView() {
        return edgeView;
    }

    public TreeView<Renderable> getVertexView() {
        return vertexView;
    }

    public Button getCheckButton() {
        return mainToolBar.getCheckButton();
    }

    public ToggleButton getGridButton() {
        return mainToolBar.getGridButton();
    }

    public Group getMeshGroup() {
        return meshGroup;
    }

    public ToggleButton getCullingButton() {
        return mainToolBar.getCullingButton();
    }

    public Button getWriteReportButton() {
        return mainToolBar.getWriteReportButton();
    }

    public TreeView<Renderable> getErrorTree() {
        return errorView;
    }

    public Stage getMainStage() {
        return stage;
    }

    public void unselectEverything() {
        buildingsView.getSelectionModel().clearSelection();
        vegetationView.getSelectionModel().clearSelection();
        transView.getSelectionModel().clearSelection();
        tunnelView.getSelectionModel().clearSelection();
        waterView.getSelectionModel().clearSelection();
        terrainView.getSelectionModel().clearSelection();
        cityFurnitureView.getSelectionModel().clearSelection();
        otherObjectsView.getSelectionModel().clearSelection();
        bridgeView.getSelectionModel().clearSelection();
        polygonView.getSelectionModel().clearSelection();
        edgeView.getSelectionModel().clearSelection();
        vertexView.getSelectionModel().clearSelection();
        attributeView.getSelectionModel().clearSelection();
        errorView.getSelectionModel().clearSelection();
    }

    public ToggleButton getLod1Btn() {
        return mainToolBar.getLod1Btn();
    }

    public ToggleButton getLod2Btn() {
        return mainToolBar.getLod2Btn();
    }

    public ToggleButton getLod3Btn() {
        return mainToolBar.getLod3Btn();
    }

    public ToggleButton getLod4Btn() {
        return mainToolBar.getLod4Btn();
    }

    public Button getWorldBtn() {
        return mainToolBar.getWorldBtn();
    }

    public Button getResetCameraBtn() {
        return mainToolBar.getResetCameraBtn();
    }

    public ToggleButton getHideRoofBtn() {
        return mainToolBar.getHideRoofBtn();
    }

    public ImageView getNorthArrow() {
        return northArrow;
    }

    public Button getSaveBtn() {
        return mainToolBar.getSaveBtn();
    }

    public SplitPane getMainContainer() {
        return mainContainer;
    }

    public ListView<String> getGlobalErrorsView() {
        return globalErrorsView;
    }

    public void takeViewScreenshot() throws IOException {
        WritableImage snapshot = geomScene.snapshot(null, null);
        File outputFile = new File("img.png");
        BufferedImage bImage = SwingFXUtils.fromFXImage(snapshot, null);
        ImageIO.write(bImage, "png", outputFile);
    }

    public void zoomOutForBoundingBox(BoundingBox b) {
        double longestSide = b.getDiagonalLength() * 0.4;
        double d = longestSide / Math.tan(Math.toRadians(30) / 2);
        translateZ = -d;
        camera.setTranslateZ(translateZ);
        camera.setTranslateX(0);
        camera.setTranslateY(0);
        highlightController.changeScaling(-translateZ);
        controller.setOriginBB(b);
    }

    public void showCityObjectDetail(CityObject co) {
        FeatureType coType = co.getFeatureType();
        TreeView<Renderable> featureView;
        Tab featureTab;
        if (coType == FeatureType.BUILDING){
            featureView = buildingsView;
            featureTab = buildingsTab;
        } else if (coType == FeatureType.VEGETATION){
            featureView = vegetationView;
            featureTab = vegetationTab;
        } else if (coType == FeatureType.TRANSPORTATION){
            featureView = transView;
            featureTab = transportationTab;
        } else if (coType == FeatureType.TUNNEL){
            featureView = tunnelView;
            featureTab = tunnelTab;
        } else if (coType == FeatureType.BRIDGE){
            featureView = bridgeView;
            featureTab = bridgeTab;
        } else if (coType == FeatureType.WATER){
            featureView = waterView;
            featureTab = waterTab;
        } else if (coType == FeatureType.LAND){
            featureView = terrainView;
            featureTab = terrainTab;
        } else if (coType == FeatureType.CITY_FURNITURE){
            featureView = cityFurnitureView;
            featureTab = cityFurnitureTab;
        } else if (coType == FeatureType.GENERIC_CITY_OBJECT){
            featureView = otherObjectsView;
            featureTab = otherObjectsTab;
        } else {
            throw new IllegalArgumentException("unknown feature type: " + coType);
        }
        featurePane.getSelectionModel().select(featureTab);
        searchField.setText(co.getGmlId().getGmlString());
        searchBtn.fire();
        renderer.render(co);
    }

    public MainToolBar getMainToolbar() {
        return mainToolBar;
    }

    public void clearHighlights() {
        highlightController.clearHighlights();
    }

    public CityDoctorController getController() {
        return controller;
    }

    public Button getOpenBtn() {
        return mainToolBar.getOpenBtn();
    }

    public VertexClickHandler getClickHandler() {
        return clickHandler;
    }

    public FeatureType getSelectedTab() {
        return selectedTab;
    }
}
